/********************************************************************/
/*                       MasterMind v. 0.001                        */
/*                   Sfida 2000 - it.fan.dewdney                    */
/********************************************************************/

/********************************************************************/
/* Versione di                                                      */
/* Roberto Corda (analisi, implementazione)                         */
/* Doppia G. (analisi,supporto morale, *non c'ho il compilatore*)   */
/********************************************************************/

/* descrizione dell' algoritmo(i)
   -la prima puntata e' puramente casuale
   -dopo di che si prova una delle puntate ancora possibili
   - la scelta tra quelle possibili puo' essere fatta:
     * a caso (simpleguess)
     * scegliendo quella che minimizza la quntita' massima di
       puntate possibili rimaste (minimaxguess)
     * scegliendo quella che minimizza la media delle puntate
       (statguess)
     * quella che ha la migliore natura-budda (zenguess)
   abbiamo implementato tutte e 4 le strategie (simpleguess, minimaxguess,
   statguess e zenguess) e per ragioni sperimentali abbiamo scelto
   di utilizzare zenguess.
   Consigliamo di cercare questa funzione e notare la sua particolare
   e sofisticata implementazione.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define byte    unsigned char
#define MAX     5

/* INIZIO define variabili e prototipi funzioni aggiunte */

#define HOLESNUM  5
#define COLORNUM 10
#define MAXTRIES 100
#define NOTACOLORA 42
#define NOTACOLORB 66;
#define BLACK 0
#define WHITE 1
#define COMBNUM (10*10*10*10*10)
#define RESNUM (20)

struct tcomb {
 int val[5];
};

struct tres {
 int val[2];
};

int allcat[7][5] =
{
 {2,1,1,1,0},
 {1,1,1,1,1},
 {2,2,1,0,0},
 {3,1,1,0,0},
 {3,2,0,0,0},
 {4,1,0,0,0},
 {5,0,0,0,0}
};


struct tcomb combord[COMBNUM];

int catgrpnum[7];

struct tcomb comb[COMBNUM];
struct tres  res[RESNUM];
struct tcomb tries[MAXTRIES];
int checks[MAXTRIES][2];
int left;

struct tcomb simpleguess(int ntry);
struct tcomb minimaxguess(int ntry);
struct tcomb statguess(int ntry);
struct tcomb zenguess(int ntry);
struct tcomb minimaxguess_par(int ntry,int simple, int level, int cat);
struct tcomb statguess_par(int ntry,int simple, int level, int cat);
/* FINE define variabili e funzioni aggiunte */

void Punta (byte *punt, byte N, byte B, byte t);
void Verifica (byte *num, byte *punt, byte *N, byte *B);
void Output (byte *punt, byte N, byte B);

void Output (byte *punt, byte N, byte B) {
    int i;
    for (i = 0; i < MAX; i++)
        printf ("%d", punt[i]);
    printf ("   ---->   ");
    printf ("(%d - %d)\n", N, B);
}

void Verifica (byte *num, byte *punt, byte *N, byte *B) {
    int i, j;
    byte a[MAX], b[MAX];
    
    *B = *N = 0;
    /* cerca gli N, ovvero i numeri al posto giusto */
    for (i = 0; i < MAX; i++) {
        if (punt[i] == num[i]) {
            ++ *N;
            a[i] = b[i] = 0;
        } else
            a[i] = b[i] = 1;
    }
    /* cerca i B, cioe' i numeri al posto sbagliato */
    for (i = 0; i < MAX; i++)
        for (j = 0; j < MAX; j++)
            /* i != j e' gia' compreso */
            if (a[i] && b[j] && (punt[i] == num[j]) ) {
                a[i] = b[j] = 0;
                ++ *B;
            }
}

void Punta (byte *punt, byte N, byte B, byte t) {
    
   struct tcomb guess;
   int i;

   /* NB i numeri di tentativo partono da 1 e non da 0 */
   /* memorizzo i risultati ottenuti nella puntata precedente */
   if (t > 1) {
     checks[t-2][BLACK] = N;
     checks[t-2][WHITE] = B;
   }

   guess = zenguess(t-1);

   tries[t-1] = guess;
   punt[0] = guess.val[0];
   punt[1] = guess.val[1];
   punt[2] = guess.val[2];
   punt[3] = guess.val[3];
   punt[4] = guess.val[4];

}


/* wrapper di minimax_par */
struct tcomb zenguess(int ntry) {

 return minimaxguess_par(ntry,0,10000,0);
 //return statguess_par(ntry,0,10000,0);
}


/* INIZIO funzioni aggiunte */

void swapcomb(int a, int b){
 struct tcomb buff;
 buff = comb[a];
 comb[a] = comb[b];
 comb[b] = buff;
}

/* algoritmo semplice */
/* prova ogni volta una combinazione tra quelle possibili rimaste */
struct tcomb simpleguess(int ntry){
 int r,black,white,newleft,i;

 if ( ntry == 0 ) {
  left = COMBNUM;
  return( comb[rand()%COMBNUM] );
  }

 for (i=0,newleft=0;i<left;i++){
   check(comb[i],tries[ntry-1],&black,&white);
   if ((checks[ntry-1][BLACK] == black) &&
      (checks[ntry-1][WHITE] == white))
      swapcomb(newleft++,i);
 }

 left = newleft;
 return( comb[rand()%left] );
}

/* algoritmo stile minimax */
/* prova ogni volta una combinazione tra quelle possibili rimaste */
/* scegliendo quella che minimizza il caso peggiore */
struct tcomb minimaxguess(int ntry){
 int r,black,white,newleft,i;
 int outpos, pos, max, min, val;

 /* se e' la prima giocata */
 if ( ntry == 0 ) {
  left = COMBNUM;
  return( comb[rand()%COMBNUM] );
  }

 /* se e' una delle giocate successive */

 /* elimina giocate impossibili */
 for (i=0,newleft=0;i<left;i++){
   check(comb[i],tries[ntry-1],&black,&white);
   if ((checks[ntry-1][BLACK] == black) &&
      (checks[ntry-1][WHITE] == white))
      swapcomb(newleft++,i);
 }
 left = newleft;

 /* scegli la prossima giocata che minimizza il caso peggiore */
 min = COMBNUM + 10;
 outpos = 0;
 /* per ognuna delle rimanenti combinazioni possibili */
 for (pos=0;pos<left;pos++)
 {
   /* cerco il caso peggiore con questa combinazione */
   max = 0;
   /* per ognuno dei risultati possibili */
   for (r=0;r<RESNUM;r++)
   {
     // conta quante possibilita' rimarrebbero con quella puntata
     // e quel risultato (comb[pos], res[r]
     val = 0;
     for (i=0;i<left;i++)
     {
      check(comb[i],comb[pos],&black,&white);
      if ((res[r].val[BLACK] == black) &&
        (res[r].val[WHITE] == white))
        val++;

      // prende il massimo
      // fix se sono uguali?
      if (val > max)
        max = val;
     }
   }
   // prende il minimax
   // fix se sono uguali?
   if (max < min) {
      min = max;
      outpos = pos;
   }
 }

 return( comb[outpos] );
}


struct tcomb putcomb(struct tcomb *cmb,int c0,int c1,int c2,int c3,int c4){
 cmb->val[0] = c0;
 cmb->val[1] = c1;
 cmb->val[2] = c2;
 cmb->val[3] = c3;
 cmb->val[4] = c4;

 return *cmb;
}

struct tcomb makecomb(int c0,int c1,int c2,int c3,int c4){
 struct tcomb cmb;

 cmb.val[0] = c0;
 cmb.val[1] = c1;
 cmb.val[2] = c2;
 cmb.val[3] = c3;
 cmb.val[4] = c4;

 return cmb;
}

int printguess(struct tcomb cmb){
 printf("|%d%d%d%d%d|",
   cmb.val[0],cmb.val[1],cmb.val[2],cmb.val[3],cmb.val[4]);
}

int isin(int v, struct tcomb comb){
 int i;

 for(i=0;i<HOLESNUM;i++)
  if (comb.val[i] == v)
    return i;

 return -1;
}


int maxvect(int *vect) {
  int i;
  int max;
  int pos;

  pos = -1;
  max = 0;
  for (i=0; i<10; i++)
  {
    if (vect[i] > max)
    {
      pos = i;
      max = vect[i];
    }
  }
  return pos;
}

/* presa una combinazione stabilisce a che categoria appartiene */
int categoria(struct tcomb guess) {
  int grpcol[10];
  int cat[5];
  int pos, gotcha;

  int i,j,out;

  /* azzera */
  for (i=0; i<10; i++) grpcol[i] = 0;

  /* conta quanti per ogni colore */
  for (i=0; i<5; i++) {
   grpcol[guess.val[i]]++;
  }

  /* ordina in numeri di colori */
  for (i=0; i<5;i++) cat[i]=0;
  for (i=0; i<5;i++) {
   pos = maxvect(grpcol);
   if (pos == -1)
     break;
   cat[i] = grpcol[pos];
   grpcol[pos] = 0;
  }

  /* cerca la categoria */
  gotcha = 0;
  for (i=0; (i<7) && (!gotcha); i++) {
    for (j=0; j<5; j++) {
      if (cat[j] != allcat[i][j]) {
          gotcha = 0;
          break;
        }
      else {
        gotcha = 1;
        out = i;
      }
    }
  }

  if (gotcha)
    return out;
  else {
    printf("ERRORE: categoria inesistente");
    exit(0);
  }

}

// riempie tutti i possibili risultati

int init(){
 int h0,h1,h2,h3,h4,i,j,n,c;

 for (h4 = 0; h4<COLORNUM; h4++)
 for (h3 = 0; h3<COLORNUM; h3++)
 for (h2 = 0; h2<COLORNUM; h2++)
 for (h1 = 0; h1<COLORNUM; h1++)
 for (h0 = 0; h0<COLORNUM; h0++)
  {
   i = h0 +
       h1*COLORNUM +
       h2*COLORNUM*COLORNUM +
       h3*COLORNUM*COLORNUM*COLORNUM +
       h4*COLORNUM*COLORNUM*COLORNUM*COLORNUM;

   comb[i].val[0] = h0;
   comb[i].val[1] = h1;
   comb[i].val[2] = h2;
   comb[i].val[3] = h3;
   comb[i].val[4] = h4;
   //printf("%d%d%d%d%d,",h4,h3,h2,h1,h0);
  }

  n = 0;
  for (h0 = 0; h0<=5; h0++)
  for (h1 = 0; h1+h0 <= 5; h1++)
  {

   res[n].val[BLACK] = h0;
   res[n].val[WHITE] = h1;

   n++;
  }

  // mette in combord le combinazioni ordinate per categoria
 n=0 ;
 for (i=0; i<7; i++) catgrpnum[i] = 0;
 for (j = 0; j<7; j++) {
   for (i =0; i<100000; i++) {
     c = categoria(comb[i]);
     if (c == j) {
       combord[n] = comb[i];
       n++;
       catgrpnum[c]++;
     }
   }
 }
}

/* conta quanti pallini bianchi e' neri sono stati totalizzati */
/* dalla prova guess */
int check(struct tcomb right, struct tcomb guess,int *black, int *white) {
  int i,w;


  *black = 0;
  *white = 0;

  for (i=0;i<HOLESNUM;i++){
   if (right.val[i]==guess.val[i]) {
    (*black)++;
    right.val[i] = NOTACOLORA;
    guess.val[i] = NOTACOLORB;
   }
  }

  for (i=0;i<HOLESNUM;i++){
   if ((w = isin(guess.val[i],right)) != -1) {
    (*white)++;
    right.val[w] = NOTACOLORA;
   }
  }

}

int min_of(int a, int b) {

 if (a<b)
   return a;
 else
   return b;
}

struct tcomb simpleguess_check(int ntry, int categoria){
 int r,black,white,newleft,i;
 int offset;

 if ( ntry == 0 ) {
  left = COMBNUM;

  /* scelta a random solo tra quelli della categoria */
  offset = 0;
  for (i =0; i<categoria; i++) offset += catgrpnum[i];
  return( combord[offset + rand()%catgrpnum[categoria]] );
  }

 for (i=0,newleft=0;i<left;i++){
   check(comb[i],tries[ntry-1],&black,&white);
   if ((checks[ntry-1][BLACK] == black) &&
      (checks[ntry-1][WHITE] == white))
      swapcomb(newleft++,i);
 }

 left = newleft;
 return( comb[rand()%left] );
}


struct tcomb minimaxguess_par(int ntry,int simple, int level, int cat){
 int r,black,white,newleft,i;
 int outpos, pos, max, min, val,check_pos, rnd_pos;

 // nelle prime tre giocate usa simple guess
 // di una certa categoria se cat > 0 altrimenti sceglie tra tute le giocate
 if (ntry <= simple)
 {
  if (cat >= 0)
    return simpleguess_check(ntry,cat);
  else
    return simpleguess(ntry);

 }
 /* se e' una delle giocate successive */

 /* elimina giocate impossibili */
 for (i=0,newleft=0;i<left;i++){
   check(comb[i],tries[ntry-1],&black,&white);
   if ((checks[ntry-1][BLACK] == black) &&
      (checks[ntry-1][WHITE] == white))
      swapcomb(newleft++,i);
 }
 left = newleft;

 /* scegli la prossima giocata che minimizza il caso peggiore */
 min = COMBNUM + 10;
 outpos = 0;

 /* un numero casuale di elementi tra cui scegliere */
 check_pos = min_of(1+(level/left),left);

 /* sceglie check_pos combinazioni a caso e le mette all'inizio di comb*/
 for (i=0; i<check_pos; i++) {

    rnd_pos = rand()%left;
    swapcomb(rnd_pos,i);
 }

 /* per ognuna delle rimanenti combinazioni possibili */
 for (pos=0;pos<check_pos;pos++)
 {

   /* cerco il caso peggiore con questa combinazione */
   max = 0;
   /* per ognuno dei risultati possibili */
   for (r=0;r<RESNUM;r++)
   {
     // conta quante possibilita' rimarrebbero con quella puntata
     // e quel risultato (comb[pos], res[r]
     val = 0;
     for (i=0;i<left;i++)
     {
      check(comb[i],comb[pos],&black,&white);
      if ((res[r].val[BLACK] == black) &&
        (res[r].val[WHITE] == white))
        val++;

      // prende il massimo
      // fix se sono uguali?
      if (val > max)
        max = val;
     }
   }
   // prende il minimax
   // fix se sono uguali?
   if (max < min) {
      min = max;
      outpos = pos;
   }
 }

 return( comb[outpos] );
}

struct tcomb statguess_par(int ntry,int simple, int level, int cat){
 int r,black,white,newleft,i;
 int outpos, pos, min, val, check_pos, rnd_pos;
 long media;

 // nelle prime tre giocate usa simple guess
 // di una certa categoria se cat > 0 altrimenti sceglie tra tute le giocate
 if (ntry <= simple)
 {
  if (cat >= 0)
    return simpleguess_check(ntry,cat);
  else
    return simpleguess(ntry);

 }
 /* se e' una delle giocate successive */

 /* elimina giocate impossibili */
 for (i=0,newleft=0;i<left;i++){
   check(comb[i],tries[ntry-1],&black,&white);
   if ((checks[ntry-1][BLACK] == black) &&
      (checks[ntry-1][WHITE] == white))
      swapcomb(newleft++,i);
 }
 left = newleft;

 /* scegli la prossima giocata che minimizza la media dei casi*/

 min = 1000000;
 outpos = 0;

 /* un numero casuale di elementi tra cui scegliere */
 check_pos = min_of(1+(level/left),left);

 /* sceglie check_pos combinazioni a caso e le mette all'inizio di comb*/
 for (i=0; i<check_pos; i++) {

    rnd_pos = rand()%left;
    swapcomb(rnd_pos,i);
 }

 /* per ognuna delle rimanenti combinazioni possibili */
 for (pos=0;pos<check_pos;pos++)
 {

   /* sum(rimanenti^2) */
   media = 0;
   /* per ognuno dei risultati possibili */
   for (r=0;r<RESNUM;r++)
   {
     // conta quante possibilita' rimarrebbero con quella puntata
     // e quel risultato (comb[pos], res[r]
     val = 0;
     for (i=0;i<left;i++)
     {
      check(comb[i],comb[pos],&black,&white);
      if ((res[r].val[BLACK] == black) &&
        (res[r].val[WHITE] == white))
        val++;
     }
     media += val*val;
   }
   // prende il minimax
   if (pos == 0)
     min = media;
   else
     if (media < min) {
      min = media;
      outpos = pos;
     }
 }

 return( comb[outpos] );
}

/* FINE funzioni aggiunte */

int main (int argv, char **argc) {
    byte num[MAX],      /* numero */
         punt[MAX],     /* puntata */
         N, B,          /* Neri -> posto giusto 
                           Bianchi -> posto sbagliato */
         t;             /* tempo */
         int tot_t = 0, /* tentativi totali */
                 n_risolti = 0; /* numeri risolti */
    FILE *f;            /* puntatore al file */
    char fpath[128];    /* percorso\nome file */
    int i;

    /************************************************/
    /* funzione aggiunta in main (inizializzazione) */
    /************************************************/
    init();
    /************************************************/


    /* se ci sono argomenti, assume il primo parametro come 
       nome del file di input, che altrimenti e' mind.txt */
    if (argv > 1)
        strcpy (fpath, argc[1]);
    else
        strcpy (fpath, "mind.txt");
    if ((f = fopen (fpath, "r")) == NULL) {
        fprintf (stderr, "Errore: impossibile aprire il file %s\n", fpath);
        exit (1);
    }

    printf ("File di input: %s\n", fpath);

    /* ripete per ogni riga del file */
    while ((fscanf (f, "%c%c%c%c%c\n", &num[0], &num[1], &num[2], &num[3], &num[4]) == 5)) {
        for (i = 0; i < MAX; i++)
            num[i] -= 48;
        
        /* ripristina le condizioni iniziali */
        B = N = t = 0;
         
        /* ripete fino alla vittoria */
        do { 
            Punta (punt, N, B, ++t);
            Verifica (num, punt, &N, &B);
            Output (punt, N, B);
        } while (N < 5);
        n_risolti++;
        tot_t +=t;
        printf ("Risolti: %d; Mosse: %d; In totale: %d\n", n_risolti,t,tot_t);
    }
    fclose (f);
}
